from datetime import datetime
from typing import TypeVar, Iterator, Optional, Callable

import numpy as np
from numpy.typing import NDArray
from scipy.constants import mu_0

from scipy.sparse import csc_matrix
from scipy.sparse.linalg import factorized

from cfsem import gs_operator_order4, flux_circular_filament


T = TypeVar("T")

# Finite-difference coeffs
# https://en.wikipedia.org/wiki/Finite_difference_coefficient
_DDX_CENTRAL_ORDER4 = np.array(
    [
        # 4th-order central difference for first derivative
        (-2, 1 / 12),
        (-1, -2 / 3),
        # (0, 0.0),
        (1, 2 / 3),
        (2, -1 / 12),
    ]
)

_DDX_FWD_ORDER4 = np.array(
    [
        # 4th-order forward difference for first derivative
        (0, -25 / 12),
        (1, 4),
        (2, -3),
        (3, 4 / 3),
        (4, -1 / 4),
    ]
)

_DDX_BWD_ORDER4 = -_DDX_FWD_ORDER4  # Reverse & flip signs


def _progressbar(it: list[T], suffix="", show_every: int = 1) -> Iterator[T]:
    """A simple terminal progressbar."""
    size = 30
    count = len(it)

    start = datetime.now()

    def show(j):
        x = int(size * j / count)
        print(
            "[{}{}] {}/{} {}, Elapsed: {:.3f} [s]".format(
                "#" * x,
                "." * (size - x),
                j,
                count,
                suffix,
                (datetime.now() - start).total_seconds(),
            ),
            end="\r",
            flush=True,
        )

    show(0)
    for i, item in enumerate(it):
        yield item
        if (i % show_every == 0) or (i == count - 1):
            show(i + 1)
    print("\n", flush=True)


def gradient_order4(
    z: NDArray, xmesh: NDArray, ymesh: NDArray
) -> tuple[NDArray, NDArray]:
    """
    Calculate gradient by 4th-order finite difference.

    `numpy.gradient` exists and is fast and convenient, but only uses a second-order difference,
    which produces unacceptable error in B-fields (well over 1% for typical geometries).

    ## Errors

        * If the input grids are not regular
        * If any input grid dimensions have size less than 5

    ## References

        * [1] “Finite difference coefficient,” Wikipedia. Aug. 22, 2023.
              Accessed: Mar. 29, 2024. [Online].
              Available: https://en.wikipedia.org/w/index.php?title=Finite_difference_coefficient

    Args:
        z: [<xunits>] 2D array of values on which to calculate the gradient
        xmesh: [m] 2D array of coordinates of first dimension
        ymesh: [m] 2D array of coordinates of second dimension

    Returns:
        (dzdx, dzdy) [<xunits>/m] 2D arrays of gradient components
    """
    nx, ny = z.shape
    dx = xmesh[1][0] - xmesh[0][0]
    dy = ymesh[0][1] - ymesh[0][0]

    # Check regular grid assumption
    assert np.all(
        np.abs(np.diff(xmesh[:, 0]) - dx) / dx < 1e-6
    ), "This method is only implemented for a regular grid"
    assert np.all(
        np.abs(np.diff(ymesh[0, :]) - dy) / dy < 1e-6
    ), "This method is only implemented for a regular grid"

    dzdx = np.zeros_like(z)
    for offs, w in _DDX_CENTRAL_ORDER4:
        start = int(2 + offs)
        end = int(nx - 2 + offs)
        dzdx[2:-2, :] += (
            w * z[start:end, :] / dx
        )  # Central difference on interior points
    for offs, w in _DDX_FWD_ORDER4:
        offs = int(offs)
        dzdx[0:2, :] += (
            w * z[offs : offs + 2, :] / dx
        )  # One-sided difference on left side
    for offs, w in _DDX_BWD_ORDER4:
        start = int(-2 + offs)
        end = int(nx + offs)
        dzdx[-2:, :] += w * z[start:end, :] / dx  # right side

    dzdy = np.zeros_like(z)
    for offs, w in _DDX_CENTRAL_ORDER4:
        start = int(2 + offs)
        end = int(ny - 2 + offs)
        dzdy[:, 2:-2] += w * z[:, start:end] / dy  # Interior points
    for offs, w in _DDX_FWD_ORDER4:
        offs = int(offs)
        dzdy[:, 0:2] += w * z[:, offs : offs + 2] / dy  # One-sided difference on bottom
    for offs, w in _DDX_BWD_ORDER4:
        start = int(-2 + offs)
        end = int(ny + offs)
        dzdy[:, -2:] += w * z[:, start:end] / dy  # top

    return dzdx, dzdy


def calc_flux_density_from_flux(
    psi: NDArray, rmesh: NDArray, zmesh: NDArray
) -> tuple[NDArray, NDArray]:
    """
    Back-calculate B-field from poloidal flux per Wesson eqn 3.2.2 by 4th-order finite difference,
    modified to use total poloidal flux instead of flux per radian.

    This avoids an expensive sum over filamentized contributions at the expense of some numerical error.

    # Errors

        * If the input grids are not regular
        * If any input grid dimensions have size less than 5

    # References

        * [1] J. Wesson, Tokamaks. Oxford, New York: Clarendon Press, 1987.

    Args:
        psi: [Wb] poloidal flux
        rmesh: [m] 2D r-coordinates
        zmesh: [m] 2D z-coordinates

    Returns:
        (br, bz) [T] 2D arrays of poloidal flux density
    """

    dpsidr, dpsidz = gradient_order4(psi, rmesh, zmesh)

    r_inv = rmesh**-1

    br = -r_inv * dpsidz / (2.0 * np.pi)  # [T]
    bz = r_inv * dpsidr / (2.0 * np.pi)  # [T]

    return (br, bz)


def flux_solver(grids: tuple[NDArray, NDArray]) -> Callable[[NDArray], NDArray]:
    """
    Linear solver for extracting a flux field from a toroidal current density distribution
    using a 4th-order finite difference approximation of the Grad-Shafranov PDE.
    For `jtor` toroidal current density shaped like (nr, nz), call like `psi = flux_solver(rhs)`
    to get `psi` in [Wb] or [V-s], where `rhs = -2.0 * np.pi * mu_0 * rmesh * jtor` with the boundary
    values set to the circular-filament solved flux.

    Args:
        grids: [m] regular 1D r,z grids

    Returns:
        solver: factorized solver for Grad-Shafranov differential operator
    """
    # Build Grad-Shafranov Delta* linear operator for finite difference
    # as a sparse matrix
    _ = _check_regular(grids)
    rgrid, zgrid = grids
    nr = rgrid.size
    nz = zgrid.size
    vals, rows, cols = gs_operator_order4(*grids)
    operator = csc_matrix((vals, (rows, cols)), shape=(nr * nz, nr * nz))
    # Store LU factorization of operator matrix to allow fast, reusable
    # solves using different right-hand-side (different current density)
    return factorized(operator)


def solve_flux_axisymmetric(
    grids: tuple[NDArray, NDArray],
    meshes: tuple[NDArray, NDArray],
    current_density: NDArray,
    solver: Optional[Callable[[NDArray], NDArray]] = None,
) -> NDArray:
    """
    Calculate the flux field associated with a given toroidal current density distribution,
    by solving the Grad-Shafranov PDE.

    This calculation is most commonly used for the plasma, but is in fact more general,
    and applies to anything with an equivalent toroidal current density and axisymmetry.

    Args:
        grids: [m] 1D r,z regular coordinate grids
        meshes: [m] 2D meshgrids made from grids like np.meshgrid(*grids, indexing="ij")
        current_density: [A/m^2], shape (nr, nz), toroidal current density on finite-difference mesh
        solver: Optionally, provide a pre-initialized linear solver. See `cfsem.utils.flux_solver`.

    Returns:
        poloidal flux field, [Wb] with shape (nr, nz)
    """
    # Build the differential operator, if needed
    solver = solver or flux_solver(grids)

    # Unpack and filter down to just useful inputs
    dr, dz = _check_regular(grids)  # [m] grid spacing
    area = dr * dz  # [m^2]
    rmesh, zmesh = meshes  # [m]
    nonzero_inds = np.where(current_density != 0.0)
    current_density_nonzero = np.ascontiguousarray(
        current_density[nonzero_inds]
    )  # [A/m^2]
    rmesh_nonzero = np.ascontiguousarray(rmesh[nonzero_inds])  # [m]
    zmesh_nonzero = np.ascontiguousarray(zmesh[nonzero_inds])  # [m]
    # Solve `Delta* @ psi = -mu_0 * 2pi * rmesh * jtor`
    #   Set up right-hand-side of Grad-Shafranov
    rhs = -(2.0 * np.pi * mu_0) * rmesh * current_density  # [Wb/m^2]
    #   Set flux boundary condition
    #   For most relevant grid sizes (up to 500 X 500), doing the O(N^3)
    #   circular-filament flux calc is faster than the linear solve
    #   and therefore faster than doing an extra fixed-boundary linear solve
    #   in order to use Von Hagenow's asymptotically-O(N^2logN) method.
    ifil = (area * current_density_nonzero).flatten()  # [A] plasma filament current
    rfil = rmesh_nonzero.flatten()
    zfil = zmesh_nonzero.flatten()
    for s in [[0, ...], [-1, ...], [..., 0], [..., -1]]:  # All boundary slices
        rhs[s[0], s[1]] = flux_circular_filament(
            ifil, rfil, zfil, rmesh[s[0], s[1]], zmesh[s[0], s[1]]
        )
    #   Do the actual linear solve
    psi = solver(rhs.flatten()).reshape(rmesh.shape)  # [Wb]

    return psi


def _check_regular(grids: tuple[NDArray, NDArray], tol=1e-6) -> tuple[float, float]:
    """Check that grids are regular and returns spacing"""
    rgrid, zgrid = grids
    drs = np.diff(rgrid)
    dzs = np.diff(zgrid)
    drmean = float(np.mean(drs))
    dzmean = float(np.mean(dzs))
    assert np.all(np.abs(drs - drmean) / drmean < 1e-4), "Grids must be regular"
    assert np.all(np.abs(dzs - dzmean) / dzmean < 1e-4), "Grids must be regular"

    return drmean, dzmean  # [m]


def _rect_mask(
    meshes: tuple[NDArray, NDArray],
    extent: tuple[float, float, float, float],
    pad: tuple[float, float] = (0.0, 0.0),
) -> tuple[NDArray, tuple[NDArray[np.intp], ...]]:
    """
    Get a boolean mask and interior indices of the rectangular region
    of an R-Z mesh spanned by an extent with padding.

    Args:
        meshes: [m] 2D r,z meshgrids
        extent: [m] rmin, rmax, zmin, zmax extent of region to mask
        pad: [m] r,z padding to add on either side of the extent. Defaults to (0.0, 0.0).

    Returns:
        2D mask, interior indices
    """
    rmin, rmax, zmin, zmax = extent  # all [m]
    rpad, zpad = pad
    rmesh, zmesh = meshes

    # Apply padding
    rmin, rmax, zmin, zmax = rmin - rpad, rmax + rpad, zmin - zpad, zmax + zpad

    # Build mask
    mask = np.ones_like(rmesh)
    mask *= np.where(rmesh >= rmin, True, False)
    mask *= np.where(rmesh <= rmax, True, False)
    mask *= np.where(zmesh >= zmin, True, False)
    mask *= np.where(zmesh <= zmax, True, False)
    inds = np.where(mask > 0.0)

    return mask, inds


def _pad_extent(
    extent: tuple[float, float, float, float], pad: tuple[float, float]
) -> tuple[float, float, float, float]:
    """Add r,z padding to both sides of an extent"""
    rmin, rmax, zmin, zmax = extent
    rpad, zpad = pad
    return rmin - rpad, rmax + rpad, zmin - zpad, zmax + zpad


def _join_extents(
    a: tuple[float, float, float, float], b: tuple[float, float, float, float]
) -> tuple[float, float, float, float]:
    """The union of two rmin, rmax, zmin, zmax extents"""
    rmina, rmaxa, zmina, zmaxa = a
    rminb, rmaxb, zminb, zmaxb = b
    return min(rmina, rminb), max(rmaxa, rmaxb), min(zmina, zminb), max(zmaxa, zmaxb)



from pathlib import Path
from mktestdocs import check_md_file


def test_md():
    """Test python snippets in markdown files"""
    check_md_file(Path(__file__).parent / "../README.md")
    check_md_file(Path(__file__).parent / "../docs/index.md")


import numpy as np
from numpy.typing import NDArray

from shapely import Polygon

MAX_EDGE_LENGTH_M = 0.1
"""Default maximum length of an edge in the lowest-level discretization"""


def winding_number(
    path: NDArray, centroid: NDArray
) -> tuple[NDArray, NDArray, NDArray]:
    """
    Calculate winding number about x, y, and z axes, w.r.t. the filament centroid.
    This estimates the angle subtended by each segment of the coil path, which is
    2*pi radians per full revolution, and may be either positive or negative depending
    on the orientation of the coil winding path.

    The choice of sign convention is arbitrary, as these results exist to compare to each other,
    and for no other purpose.
    """

    # path = xyz  # [m]
    n = path.shape[1]

    # centroid = np.mean(path, axis=1)  # [m]
    x_winding = np.zeros(n)  # [rad]
    y_winding = np.zeros(n)
    z_winding = np.zeros(n)

    for i in range(n - 1):
        # Translate the target points to center on the centroid
        first_point = path[:, i] - centroid  # [m]
        second_point = path[:, i + 1] - centroid

        # Extract angle subtended by the two points about each axis
        cross_1_2 = np.cross(first_point, second_point)
        dot_1_2 = np.dot(first_point, second_point)
        x_winding[i + 1] = x_winding[i] + np.atan2(cross_1_2[0], dot_1_2)
        y_winding[i + 1] = y_winding[i] + np.atan2(cross_1_2[1], dot_1_2)
        z_winding[i + 1] = z_winding[i] + np.atan2(cross_1_2[2], dot_1_2)

    return x_winding, y_winding, z_winding  # [rad]


def unroundness(p: Polygon) -> float:
    """
    [dimensionless] ratio of perimeter to perimeter of a circle with the same area.
    Gives a heuristic estimate for how un-like a circle the shape is.
    """
    perimeter = p.boundary.length  # [m]
    circle_perimeter = 2.0 * np.pi * np.sqrt(p.area / np.pi)  # [m]
    return perimeter / circle_perimeter  # [dimensionless]


def poly_angle(
    p: Polygon,
    centroid: tuple[float, float],
    max_edge_length_m: float = MAX_EDGE_LENGTH_M,
) -> float:
    """
    Get peak winding number in [rad] of the polygon boundary w.r.t. the centroid.
    This is not necessarily the same as the maximum angle subtended by the polygon,
    but is a good heuristic for it, and runs in O(N) instead of O(N^2).
    """
    # Expand to 3D and segmentize to avoid missing big sections
    pr, pz = p.boundary.segmentize(max_edge_length_m).xy
    xyz = np.array([pr, np.zeros_like(pr), pz])
    centroid_arr = np.array([centroid[0], 0.0, centroid[1]])

    # Get aggregate winding number w.r.t. centroid along polygon boundary
    _wx, wy, _wz = winding_number(xyz, centroid_arr)

    return np.max(np.abs(wy))  # [rad]



"""Run each example file, failing the test on any errors"""

import os
import pathlib
import runpy

import pytest

# os.environ["DEVICE_INDUCTANCE_TESTING"] = "True"

EXAMPLES_DIR = pathlib.Path(__file__).parent / "../examples"
EXAMPLES = [
    EXAMPLES_DIR / x
    for x in os.listdir(EXAMPLES_DIR)
    if (os.path.isfile(EXAMPLES_DIR / x) and x[-3:] == ".py")
]


@pytest.mark.parametrize("example_file", EXAMPLES)
def test_example(example_file: pathlib.Path):
    """Test every example file"""
    runpy.run_path(str(example_file), run_name="__main__")



"""Tool for cutting a polyogon into pie-slice chunks"""

from itertools import product

import numpy as np
from shapely import Polygon, GeometryCollection, MultiPolygon


class RadialSlicer:
    """Tool for cutting a polyogon into pie-slice chunks"""

    centroid: tuple[float, float]
    """Centroid used to generate the slices"""

    polygons: list[Polygon]
    """Contiguous but non-overlapping slices spanning the provided extent"""

    def __init__(
        self,
        n_slices: int,
        extent: tuple[float, float, float, float],
        centroid: tuple[float, float],
    ):
        # Unpack
        self.centroid = centroid
        rmid, zmid = centroid
        centroid_arr = np.array(centroid)

        # Choose a pie radius
        #   Actualize all the points at the corners of the extent
        extent_points = [
            np.array(x) for x in product([extent[0], extent[1]], [extent[2], extent[3]])
        ]
        #   Find the largest radius from the centroid to any of the points
        extent_radii = np.linalg.norm(
            np.array([x - centroid_arr for x in extent_points]), axis=1
        )
        pie_radius = 1.05 * np.max(extent_radii)

        pie_angles = np.linspace(0.0, 2.0 * np.pi, n_slices + 1)
        pie_r = rmid + pie_radius * np.cos(pie_angles)
        pie_z = zmid + pie_radius * np.sin(pie_angles)
        pie_rz = [x for x in zip(pie_r, pie_z)]
        mids = [(rmid, zmid)] * n_slices
        #  Sections start at the midpoint, go to two points on the circle, then end back at the midpoint
        pie_sections = zip(mids, pie_rz[:-1], pie_rz[1:], mids)
        pie_slices = [Polygon(s) for s in pie_sections]

        self.polygons = pie_slices

    def slice(self, p: Polygon) -> list[Polygon]:
        """Cut polygon `p` into as many as `n_slices` new polygons or as few as 1 polygon, if no cuts are needed."""

        # Do the intersections - this can balloon into a jumble of outputs
        intersections = [p.intersection(s) for s in self.polygons]

        # Remove edge and vertex intersections
        new_polygons = [g for g in intersections if isinstance(g, Polygon)]

        for g in intersections:
            # It's possible to have one polygon intersect more than once,
            # at an edge, at a point, etc. and we have to handle all cases.
            if isinstance(g, GeometryCollection) or isinstance(g, MultiPolygon):
                new_polygons.extend([x for x in g.geoms if isinstance(x, Polygon)])

        # Remove any empty polygons; these are common
        new_polygons = [x for x in new_polygons if not x.is_empty]

        # Make sure the area adds up
        area_orig = p.area
        area_new = sum([x.area for x in new_polygons])
        assert abs(1.0 - abs(area_new / area_orig)) < 1e-3

        return new_polygons



import numpy as np

import device_inductance

from pytest import approx


from . import typical_outputs, typical_outputs_stabilized_eigenmode, typical_outputs_many_slices  # Required fixture

__all__ = ["typical_outputs", "typical_outputs_stabilized_eigenmode", "typical_outputs_many_slices"]


def test_model_reduction(
    typical_outputs: device_inductance.TypicalOutputs,
    typical_outputs_stabilized_eigenmode: device_inductance.TypicalOutputs,
    typical_outputs_many_slices: device_inductance.TypicalOutputs,
):
    """
    In general,
    dx/dt = Ax + Bu
        y = Cx + Du
    x is current vector I
    u is voltage vector V
    
    For this particular case, with `V - RI = MdI/dt`,
    the state-space reformulation is
    
    dx/dt = dI/dt = -M^-1 @ R @ I + M^-1 @ V
        y =     I = eye @ x + 0 @ u
    
    So
    `A = -M^-1 @ R`
    `B = M^-1`
    `C = eye`
    `D = 0`
    
    We don't need C and D here, because we're not going to use this
    in an actual state space formalization. In fact, the `control`
    library isn't able to simulate even 100us of the full system
    due to the presence of some extremely small timescales.
    
    We also can't easily compare to the `control` library's balred
    (balanced reduction) method here, because the coils have near zero resistance,
    so the system is _almost_ unstable under a constant input, which causes slycot
    (the fortran backend for `control.balred`) balanced reduction to
    sort-of-fail (keeps >200 modes).
    
    Instead, we can check the initial rate of change of coil current under a step in coil voltage
    between the full system and the transformed system."""

    devices = [typical_outputs.device, typical_outputs_stabilized_eigenmode.device, typical_outputs_many_slices.device]
    ndevices = len(devices)

    import matplotlib.pyplot as plt

    _fig, axes = plt.subplots(1, ndevices, sharex=True, sharey=True)

    for iax, device in enumerate(devices):
        # Unpack
        mss = typical_outputs.mss
        mcs = typical_outputs.mcs
        mcc = typical_outputs.mcc

        r_modes = device.structure_mode_resistances
        r_c = typical_outputs.r_c
        r_s = typical_outputs.r_s

        ncoils, nstructs = mcs.shape
        nmodes = device.n_structure_modes

        coils = device.coils

        # Make transformed mutual inductance matrix parts
        mss_transformed = device.structure_mode_mutual_inductances
        mcs_transformed = device.coil_structure_mode_mutual_inductances

        # Make full mutual inductance and resistance matrices
        m = np.vstack((np.hstack((mcc, mcs)), np.hstack((mcs.T, mss))))
        m_transformed = np.vstack(
            (
                np.hstack((mcc, mcs_transformed)),
                np.hstack((mcs_transformed.T, mss_transformed)),
            )
        )

        r = np.diag(np.concatenate((np.diag(r_c), np.diag(r_s)), axis=0))
        r_transformed = np.diag(
            np.concatenate((np.diag(r_c), np.diag(r_modes)), axis=0)
        )

        # Build state-space systems for voltage-to-current transfer
        m_inv = np.linalg.inv(m)
        m_transformed_inv = np.linalg.inv(m_transformed)

        a = -m_inv @ r
        b = m_inv

        a_transformed = -m_transformed_inv @ r_transformed
        b_transformed = m_transformed_inv

        # Suppose we have V = 1.0 volt I = 1.0 amp in every coil,
        # and the structure filaments start at rest and remain unforced.
        # What does each system give us for dI/dt for the coils?
        u = np.ones(ncoils + nstructs)  # [V] input voltage
        x = np.ones(ncoils + nstructs)  # [A] initial current
        u[ncoils:] = 0.0
        x[ncoils:] = 0.0
        dIdt = a @ x + b @ u

        u_transformed = np.ones(ncoils + nmodes)  # [V] input voltage
        x_transformed = np.ones(ncoils + nmodes)  # [A] initial current
        u_transformed[ncoils:] = 0.0
        x_transformed[ncoils:] = 0.0
        dIdt_transformed = a_transformed @ x_transformed + b_transformed @ u_transformed

        # The divertor coils and VS coils will have a lot of error
        # for nearly any choice of number of modes to keep - even
        # keeping >500 modes only reduces their error by about
        # a factor of 2 compared to 40 modes.
        # NOTE: Is there a physical criterion we could use to group coils instead,
        #       like comparing self inductances and de-prioritizing low-inductance coils?
        # NOTE: Consider evaluating response to each coil being charged individually
        #       in addition to a simultaneous voltage applied to all coils
        # To handle this, we can keep a looser tolerance for
        # the DIV and VS coils, sacrificing accuracy in their
        # results for the ability to simulate the rest of the system.
        atol = 25.0  # [A/s] allow a dead zone for coils with near zero response
        rtol_pf_cs = 0.001
        #   DV and VS coils are more sensitive, but should still match well since we're not truncating modes
        rtol_div_vs = 0.01
        inds_pf_cs = [
            i
            for i, c in enumerate(coils)
            if any([x in c.name.lower() for x in ["pf", "cs"]])
        ]
        inds_div_vs = [
            i
            for i, c in enumerate(coils)
            if any([x in c.name.lower() for x in ["dv", "div", "vs"]])
        ]
        inds_all = inds_pf_cs + inds_div_vs
        duplicates_exist = len(list(set(inds_all))) != len(inds_all)
        coils_missed = len(inds_all) != ncoils
        if duplicates_exist:
            # Coils shouldn't appear in both lists, but there's nothing specifically
            # preventing this from happening. If it does happen, we need to use
            # better filters.
            raise NotImplementedError(
                "Filtered coil groups need update due to duplicates"
            )
        if coils_missed:
            # Similarly, if there were coils that didn't get captured by either filter,
            # then we need to account for that category.
            raise NotImplementedError(
                "Filtered coil groups need update due to missed coils"
            )

        dIdt_pf_cs = dIdt[inds_pf_cs]
        dIdt_div_vs = dIdt[inds_div_vs]

        dIdt_transformed_pf_cs = dIdt_transformed[inds_pf_cs]
        dIdt_transformed_div_vs = dIdt_transformed[inds_div_vs]

        # We have to do a two-sided comparison because th error may be large enough
        # that the assumption that the error is small compared to the nominal value
        # does not necessarily hold.
        for i in range(len(dIdt_pf_cs)):
            assert dIdt_pf_cs[i] == approx(
                dIdt_transformed_pf_cs[i], rel=rtol_pf_cs, abs=atol
            )
        for i in range(len(dIdt_div_vs)):
            assert dIdt_div_vs[i] == approx(
                dIdt_transformed_div_vs[i], rel=rtol_div_vs, abs=atol
            )

        plt.sca(axes[iax])
        plt.scatter(
            range(ncoils),
            dIdt[:ncoils],
            color="k",
            marker="x",
            alpha=0.4,
            s=25,
            # linestyle="--",
            label="Full System",
        )
        plt.scatter(
            range(ncoils),
            dIdt_transformed[:ncoils],
            color="k",
            s=10,
            # linestyle="-",
            label="Reduced System",
        )
        plt.gca().set_yscale("log")
        plt.gca().set_xticks(range(ncoils))
        plt.gca().set_xticklabels([x.name for x in device.coils], rotation=90.0)
        plt.xlabel("Coil Number")
        plt.ylabel("dI/dt [A/s]")
        plt.title(f"{device.model_reduction_method}\n{device._n_radial_slices} slices")
        plt.legend()

    plt.show()



from dataclasses import dataclass

import numpy as np
from omas import ODS


@dataclass(frozen=True)
class CoilSeriesCircuit:
    """
    A circuit with a single power supply and one or more coils
    connected in series.
    """

    name: str
    """Human-friendly name of the circuit"""

    supply: int
    """Index of power supply for this circuit"""

    coils: list[tuple[int, float]]
    """
    Map of the coils that are in series on this circuit to their series orientation
    which should be either -1.0 or 1.0
    """

    def __post_init__(self):
        # Make sure the sign info is correctly formatted
        for i in range(len(self.coils)):
            c = self.coils[i]
            assert c[1] != 0.0, "Sign of zero will not conserve charge"
            self.coils[i] = (c[0], np.sign(c[1]))


def _extract_circuits(description: ODS) -> list[CoilSeriesCircuit]:
    # https://gafusion.github.io/omas/schema/schema_pf%20active.html
    # device_description takes the "sides" used in this schema as interleaved
    # so that the entries for each nodes are like [ps0 in, ps0 out, ...]
    # and we continue that convention here

    # Figure out how many power supplies there are
    # The connectivity info is stored as a sort of adjacency matrix,
    # with shape (nnodes, nsides) for each circuit where nsides = 2*(nsupplies + ncoils),
    # the number of inputs and outputs of circuit components.
    ncoils = len(description["pf_active.coil"].values())
    conn0_shape = description["pf_active.circuit.0"]["connections"].shape
    nsupplies = int(conn0_shape[1] / 2 - ncoils)
    ncomponents = ncoils + nsupplies
    assert (
        2 * (ncomponents) == conn0_shape[1]
    ), "Connection sides must have even dimension and match number of components"

    # For each circuit in the ODS, figure out what coils are connected
    # and in what orientation
    circuits: list[CoilSeriesCircuit] = []
    for ods_circuit in description["pf_active.circuit"].values():
        # Parse interleaved format
        name = ods_circuit["name"]
        connections = ods_circuit["connections"]
        in_connections = connections[:, :-1][:, ::2]  # Connections to "in" sides
        out_connections = connections[:, 1::2]  # Connections to "out" sides
        ps_out_connections = out_connections[:, :nsupplies]
        ps_in_connections = in_connections[:, :nsupplies]
        coil_out_connections = out_connections[:, nsupplies:]
        coil_in_connections = in_connections[:, nsupplies:]
        # First, make sure there are no parallel connections.
        # This means every "side" should be attached to no more than
        # 1 node, so we can just check the maximum value of the sum of nodes
        # connected to sides.
        assert (
            max(np.sum(connections, axis=0)) < 2
        ), f"Circuit {ods_circuit} contains parallel paths"

        # We could still have parallel paths if there are two disjoint circuits
        # represented. To check that, we can make sure there is only one power supply
        # involved in the circuit
        assert (
            np.sum(in_connections[:, :nsupplies].flatten()) == 1
        ), f"Circuit {ods_circuit} uses multiple or zero power supplies"

        assert (
            np.sum(out_connections[:, :nsupplies].flatten()) == 1
        ), f"Circuit {ods_circuit} uses multiple or zero power supplies"

        # Now that we know there are no parallel paths, we're interested in
        # which things are connected in series, and in what order.
        #     Which power supply are we using, and what are our starting and ending nodes?
        #     SIGN CONVENTION: power supplies nominally flow current from "out" to "in",
        #     while coils nominally flow current from "in" to "out" per discussion with
        #     Christoph Hasse on slack
        starting_node, ps_index = np.argwhere(ps_in_connections == 1)[0]
        ending_node, _ = np.argwhere(ps_out_connections == 1)[0]
        coils = []

        # For each node, see what coil is attached to it
        # and which side is attached. There will only be one
        # coil
        def get_next_coil(node):
            # Get the next unvisited coil attached to a given node, and its sign orientation
            visited_coil_indices = [x[0] for x in coils]
            inc = [
                np.squeeze(x)
                for x in np.argwhere(coil_in_connections[node, :])
                if x not in visited_coil_indices
            ]
            outc = [
                np.squeeze(x)
                for x in np.argwhere(coil_out_connections[node, :])
                if x not in visited_coil_indices
            ]
            if len(inc) == 1 and len(outc) == 0:
                # If we are traversing in the direction of positive current and see
                # the coil input first, it's oriented positive
                return int(inc[0]), 1.0
            elif len(inc) == 0 and len(outc) == 1:
                # If we are traversing in the direction of positive current and see
                # the coil output first, it's oriented negative
                return int(outc[0]), -1.0
            else:
                raise ValueError(
                    f"Multiple or zero unvisited coils {inc} and {outc} at node {node} "
                    "imply parallel system or open circuit"
                )

        def get_next_node(coil, sign):
            if sign == 1.0:
                return np.argwhere(coil_out_connections[:, coil] == 1)[0][0]
            elif sign == -1.0:
                return np.argwhere(coil_in_connections[:, coil] == 1)[0][0]
            else:
                raise ValueError(f"Invalid sign {sign} for coil {coil}")

        node = starting_node
        i = 0
        while node != ending_node:
            assert i <= ncoils, f"Circuit for supply {ps_index} exceeds possible length"
            coil, sign = get_next_coil(node)
            coils.append((coil, sign))
            node = get_next_node(coil, sign)
            i += 1

        circuit = CoilSeriesCircuit(
            name=name, coils=coils, supply=ps_index
        )  # Initialize output
        circuits.append(circuit)

    # Make sure all coils are represented exactly once
    all_coils_in_circuits = sum(
        [[c[0] for c in circuit.coils] for circuit in circuits], start=[]
    )
    for i in range(ncoils):
        assert i in all_coils_in_circuits, f"Coil {i} missing from circuits"
    assert (
        len(list(set(all_coils_in_circuits))) == ncoils
    ), "Extra coils or missing coils in circuits"

    return circuits



from dataclasses import dataclass
from omas import ODS
from shapely import Polygon

CONTACT_DETECTION_DISTANCE = 1e-4
"""[m] distance under which two passive structure objects are considered to be in electrical contact"""


@dataclass(frozen=True)
class PassiveStructureInput:
    """First level of structure discretization within device_inductance.
    Some discretization may have already been performed upstream."""

    parent_name: str
    """The name of the source of the info from the ODS input"""
    polygon: Polygon
    """[m] R-Z cross-section"""
    resistivity: float
    """[ohm-m] (effective) electrical resistivity of material."""


def _collect_structures(description: ODS) -> list[PassiveStructureInput]:
    """
    Combine all passive structure input geometries and resistivities into the same format,
    a polygon with a resistivity and a name.
    """

    # Collect structure polygons and resistivities
    structure_inputs: list[PassiveStructureInput] = []

    # Wall
    items = description["wall.description_2d.0.vessel"]["unit"].values()
    n_wall = 0  # Number of wall inputs
    for wall_section in items:
        for wall_elem in wall_section["element"].values():  # Segments of each wall
            n_wall += 1
            name = wall_elem["name"]
            resistivity = wall_elem["resistivity"]  # [ohm-m]
            rs = wall_elem["outline.r"]  # [m]
            zs = wall_elem["outline.z"]  # [m]
            polygon = Polygon([x for x in zip(rs, zs)])

            structure_inputs.append(PassiveStructureInput(name, polygon, resistivity))

    # pf_passive
    # These entries may or may not be wall, depending on how the wall is defined.
    for i, passive_elem in enumerate(description["pf_passive.loop"].values()):
        name = f"{passive_elem['name']}_{i}"  # Original names may not be unique
        rs = passive_elem["element.0.geometry.outline.r"]
        zs = passive_elem["element.0.geometry.outline.z"]
        resistivity = passive_elem["resistivity"]
        polygon = Polygon([x for x in zip(rs, zs)])

        structure_inputs.append(PassiveStructureInput(name, polygon, resistivity))

    # Make sure the input polygons had a valid point ordering and aren't folded up
    for inp in structure_inputs:
        assert inp.polygon.is_valid

    # Make sure all inputs are accounted
    n_pf_passive = len(description["pf_passive.loop"].values())
    assert len(structure_inputs) == n_pf_passive + n_wall, (
        "Failed to account for some structure inputs, possibly due to duplicate names of wall sections"
    )

    return structure_inputs



import numpy as np
from pytest import raises
from device_inductance.utils import _progressbar, gradient_order4


def test_progressbar():
    """Just make sure it runs and returns the input values"""
    vals = list(range(2, 7))
    bar = _progressbar(vals)
    for i, v in enumerate(bar):
        assert vals[i] == v


def test_gradient_order4():
    """Make sure order 2 and order 4 gradients converge to the same value"""

    def f(x, y):
        # Some bivariate function without important features under the grid size
        return np.sin(x / 6.0) + np.cos(y / 6.0)

    # Coarse grid
    xgrid1 = np.linspace(-5.0, 5.0, 10)
    ygrid1 = np.linspace(-7.0, 7.0, 10)
    xmesh1, ymesh1 = np.meshgrid(xgrid1, ygrid1, indexing="ij")
    z1 = f(xmesh1, ymesh1)
    grad1_order2 = np.gradient(z1, xgrid1, ygrid1)
    grad1_order4 = gradient_order4(z1, xmesh1, ymesh1)
    assert np.allclose(grad1_order2, grad1_order4, rtol=0.1)  # Related, but not great match
    with raises(AssertionError):
        # Order 2 method should not be very good at this resolution
        assert np.allclose(grad1_order2, grad1_order4, rtol=0.01)

    # Fine grid
    xgrid2 = np.linspace(-5.0, 5.0, 100)
    ygrid2 = np.linspace(-7.0, 7.0, 100)
    xmesh2, ymesh2 = np.meshgrid(xgrid2, ygrid2, indexing="ij")
    z2 = f(xmesh2, ymesh2)
    grad2_order2 = np.gradient(z2, xgrid2, ygrid2)
    grad2_order4 = gradient_order4(z2, xmesh2, ymesh2)
    assert np.allclose(grad2_order2, grad2_order4, rtol=0.01)  # Reasonably close
    with raises(AssertionError):
        # There's still significant benefit to using the order 4 method here
        assert np.allclose(grad1_order2, grad1_order4, rtol=0.005)



import numpy as np
import shapely

import device_inductance
from cfsem import (
    self_inductance_lyle6,
    flux_circular_filament,
    gs_operator_order2,
    self_inductance_distributed_axisymmetric_conductor,
)

from pytest import approx

from scipy.constants import mu_0
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import factorized

from interpn import MulticubicRectilinear

from . import typical_outputs  # Required fixture

__all__ = ["typical_outputs"]


def test_circuit_inductances(typical_outputs: device_inductance.TypicalOutputs):
    device = typical_outputs.device

    circuits = device.circuits
    coils = device.coils

    ncirc = len(circuits)

    m_interped = np.zeros((ncirc, ncirc))

    for i in range(ncirc):
        # Get a flux interpolator for this circuit
        psii = device.circuit_flux_tables[i]
        circi_psi_interp = MulticubicRectilinear.new(
            [x for x in device.grids], psii.flatten()
        )
        for j in range(ncirc):
            circj = circuits[j]

            if m_interped[i, j] != 0.0:
                # don't double up
                continue

            # Interpolate the flux from circuit `i` to all the filaments in each coil of
            # circuit `j`, applying their sign and number of turns in the process.
            #
            # Note this works fine for self-field because we're doing a smooth G-S patch
            # over each coil, so the flux field doesn't have any singularities in it
            # except possibly for the coils with windings that do not fall on a rectangular grid.
            for p, sign in circj.coils:
                coilp = coils[p]
                rs, zs, ns = coilp.rs, coilp.zs, coilp.ns
                m_interped[i, j] += np.sum(sign * ns * circi_psi_interp.eval([rs, zs]))

            m_interped[j, i] = m_interped[i, j]

    # Compare interpolated inductance to direct calc from tables
    m = device.circuit_mutual_inductances

    assert np.allclose(m, m_interped, rtol=5e-3, atol=1e-6)


def test_coil_inductances(typical_outputs: device_inductance.TypicalOutputs):
    """
    Test each coil's self inductances against the rectangular-section calc on a rectangle bounding
    the coil's filaments because we don't have the coil's nominal bounding rectangle stored in the
    ODS, and also to handle the coils that are not really an upright rectangle.

    Test mutual inductances between coils by placing a single filament with their total number of turns
    at the center of each coil, then calculating mutual inductance between each of those single-filament
    approximations. This will produce variable results for different coil shapes; in particular, the
    single-filament representation does not capture the geometry of the central solenoid well.
    """
    device = typical_outputs.device
    coils = device.coils

    # Collapse coils to single filaments
    # and estimate self-inductance as rectangular section
    coils_collapsed = {}
    for coil in coils:
        rs = [f.r for f in coil.filaments]  # [m]
        zs = [f.z for f in coil.filaments]  # [m]

        r = np.mean(rs)  # [m]
        z = np.mean(zs)  # [m]

        w = max(np.ptp(rs), 0.025)  # [m] with a minimum width and height of 2.5cm
        h = max(np.ptp(zs), 0.025)  # [m]

        n = np.sum([f.n for f in coil.filaments])  # total number of turns

        self_inductance_approx = self_inductance_lyle6(r, w, h, n)  # [H]

        coils_collapsed[coil.name] = {
            "r": r,
            "z": z,
            "n": n,
            "w": w,
            "h": h,
            "L_approx": self_inductance_approx,
        }

    # Check self inductances - because the non-rectangular ones have small aspect ratio,
    # this estimate should be quite close
    for coil in coils:
        assert np.allclose(
            [coil.self_inductance], coils_collapsed[coil.name]["L_approx"], rtol=0.2
        )

    # Check mutual inductances
    ncoils = len(coils)
    mcc_approx = np.ones((ncoils, ncoils))
    for i in range(ncoils):
        for j in range(ncoils):
            if i == j:
                mcc_approx[i, j] = coils_collapsed[coils[i].name]["L_approx"]
            else:
                coil1 = coils[i]
                coil2 = coils[j]

                ifil = np.array([1.0])
                rfil = np.array([coils_collapsed[coil1.name]["r"]])
                zfil = np.array([coils_collapsed[coil1.name]["z"]])

                rprime = np.array([coils_collapsed[coil2.name]["r"]])
                zprime = np.array([coils_collapsed[coil2.name]["z"]])

                n1 = coils_collapsed[coil1.name]["n"]
                n2 = coils_collapsed[coil2.name]["n"]
                mcc_approx[i, j] = (
                    n1
                    * n2
                    * flux_circular_filament(ifil, rfil, zfil, rprime, zprime)[0]
                )

    # Zero-out the CS columns and test the well-posed terms alone
    cs_inds = [i for i, coil in enumerate(coils) if "cs" in coil.name.lower()]
    mcc_approx_without_cs = mcc_approx.copy()
    mcc_without_cs = typical_outputs.mcc.copy()
    mcc_without_cs[:, cs_inds] = 0.0
    mcc_without_cs[cs_inds, :] = 0.0
    mcc_approx_without_cs[:, cs_inds] = 0.0
    mcc_approx_without_cs[cs_inds, :] = 0.0

    # This test only works for the coils that can reasonably be approximated
    # as a single filament, so we don't test against the CS here; they're treated the same
    # under the hood, so this still covers all the same functionality, just not every
    # available example
    assert np.allclose(mcc_without_cs, mcc_approx_without_cs, rtol=0.1, atol=1e-6)


def test_structure_self_inductances_against_grad_shafranov(
    typical_outputs: device_inductance.TypicalOutputs,
):
    """
    Test self-inductance approximation for conducting structure against an arbitrary-section
    self-inductance calc usually used for the plasma. In order to use that more detailed calc,
    we need to solve the B-field on the interior of the shape, which requires a G-S flux solve.

    To balance test rigor against time spent running tests, a very small set of filaments are tested.
    """
    rtol = 0.05
    n_to_check = 10
    structures = typical_outputs.device.structures
    nstruct = len(structures)
    # Check ~`n_to_check` instead of every entry to avoid excessive run time
    nskip = max(nstruct // n_to_check, 1)

    inductance_ratio_err_gs = []

    for _i, s in enumerate(structures[::nskip]):
        # print(i, end="\r", flush=True)
        poly = s.polygon

        # Extract the boundary of the object
        pathr, pathz = poly.exterior.xy
        boundary = (
            shapely.LineString(np.array((pathr, pathz)).T).segmentize(1e-3).xy
        )  # Discretize surface to 1mm segments
        boundary = (
            np.array([x for x in boundary[0]]),
            np.array([x for x in boundary[1]]),
        )  # Repack bad array format

        # Make a finer mesh with more padding -
        # While this does use a free boundary condition,
        # the sharp corners of the computational domain are still
        # apparent in the solution for some distance from the boundary,
        # so we need a bit of space between the boundary
        # and the object under study,
        # but need the domain to be small enough that a significant number
        # of cells are inside the object
        padding = 0.1
        extent_gs = [
            np.min(pathr) - padding,
            np.max(pathr) + padding,
            np.min(pathz) - padding,
            np.max(pathz) + padding,
        ]

        # Make a meshgrid spanning the object
        ngrid = 100
        rgrid = np.linspace(extent_gs[0], extent_gs[1], ngrid, endpoint=True)
        zgrid = np.linspace(extent_gs[2], extent_gs[3], ngrid, endpoint=True)
        rmesh, zmesh = np.meshgrid(rgrid, zgrid, indexing="ij")
        dr = rgrid[1] - rgrid[0]
        dz = zgrid[1] - zgrid[0]
        cell_area = dr * dz  # [m^2]

        # Build a normalized positive mask of the interior of the object
        # which will represent a unit current distributed more-or-less
        # evenly over the interior
        mask = np.zeros_like(rmesh)
        for ir in range(ngrid):
            for iz in range(ngrid):
                if poly.contains(shapely.Point(rgrid[ir], zgrid[iz])):
                    mask[ir, iz] = 1.0
        interior_inds = np.where(mask > 0.0)

        # Filament parameters
        ref_current = 1.0  # [A]
        nmask = mask / np.sum(mask)  # Number of turns per filament, as mask
        itor = nmask * ref_current  # [A] toroidal current filaments
        jtor = itor / cell_area  # [A/m^2]

        # Get the Grad-Shafranov operator for this mesh
        vals, rows, cols = gs_operator_order2(rgrid, zgrid)
        operator = csc_matrix(
            (vals, (rows, cols)), shape=(ngrid**2, ngrid**2), copy=True
        )

        # LU factorize the operator
        solver = factorized(operator)

        # Build the G-S system
        rhs = -mu_0 * rmesh * jtor  # [Wb/rad]

        # Set the Grad-Shafranov free boundary condition
        ifil = itor[interior_inds].flatten()  # [A]
        rfil = rmesh[interior_inds].flatten()  # [m]
        zfil = zmesh[interior_inds].flatten()  # [m]
        for ir in range(ngrid):
            for iz in [0, ngrid - 1]:
                rhs[ir, iz] = np.sum(
                    flux_circular_filament(
                        ifil, rfil, zfil, np.array(rgrid[ir]), np.array(zgrid[iz])
                    )
                    / (2.0 * np.pi)  # [Wb/rad])
                )
        for ir in [0, ngrid - 1]:
            for iz in range(ngrid):
                rhs[ir, iz] = np.sum(
                    flux_circular_filament(
                        ifil, rfil, zfil, np.array(rgrid[ir]), np.array(zgrid[iz])
                    )
                    / (2.0 * np.pi)
                )

        # Solve G-S system
        psi_gs = (
            2.0 * np.pi * solver(rhs.flatten()).reshape(rmesh.shape)
        ) / ref_current  # [Wb/A] normalized flux

        # Unlike a plasma, all the filaments carry the same current here
        # so we can approximate the self-inductance by summing over the product of
        # the number of turns and the flux.
        L_gs = np.sum(np.sum(psi_gs * nmask))

        inductance_ratio_err_gs.append(s.self_inductance / L_gs)
        assert s.self_inductance == approx(L_gs, rel=rtol)

        # We can also use the arbitrary-section inductance calc to back out the inductance
        # from the G-S solve results

        # Extract B-fields from solved flux per radian
        br, bz = device_inductance.calc_flux_density_from_flux(
            psi_gs, rmesh, zmesh
        )  # [T/A] normalized B field

        # Estimate self-inductance using distributed calc
        L_distributed, _, _ = self_inductance_distributed_axisymmetric_conductor(
            current=1.0,
            grid=(rgrid, zgrid),
            mesh=(rmesh, zmesh),
            b_part=(br, bz),
            psi_part=psi_gs,
            mask=mask,
            edge_path=boundary,
        )

        assert s.self_inductance == approx(L_distributed, rel=rtol)

    # Keeping these here because we might revisit the error plots
    # import matplotlib.pyplot as plt

    # plt.hist(inductance_ratio_err_gs)
    # n = len(inductance_ratio_err_gs)
    # plt.title(f"Relative error\nL_rectangular / L_gs\nN={n}")
    # plt.show()



from itertools import chain

from omas import ODS
from shapely import Polygon

from device_inductance.utils import _progressbar

from .input import _collect_structures
from .slicer import RadialSlicer
from .loop import PassiveStructureLoop


def _extract_structures(
    description: ODS,
    limiter: Polygon,
    n_slices: int,
    extent: tuple[float, float, float, float],
    show_prog: bool = True,
) -> list[PassiveStructureLoop]:
    """
    Extract and filamentize passive conducting structural elements
    by first meshing the 2D cross-sections to quads, then collapsing
    each quad to a singularity filament at the mean position of its vertices.

    Resistance and self-inductance are estimated by approximating each quad as
    a square of equivalent area, which may give significant error if quads of
    high aspect ratio or nonconvex shape are encountered. Because the meshing
    algorithm nominally guarantees convex elements and maximum element size is
    limited in order to keep spatial resolution, there is some hope that this
    approximation will not be too far off.

    Args:
        description: Device geometric info in the format produced by device_description
        show_prog: Display a terminal progressbar

    Returns:
        A list of passive filaments, populated with reduced geometric info and estimated self-inductances.
    """
    # Collect wall and pf_passive structures into same format
    structure_inputs = _collect_structures(description)

    # Make a slicer that will chunk inputs radially if needed (usually just for the VV),
    # prioritizing protection of the iteraction between the structures and the plasma centroid.
    slicer = RadialSlicer(n_slices, extent, (limiter.centroid.x, limiter.centroid.y))

    # Chunk, mesh, and process each input.
    # The slow part here is running gmsh on each chunk.
    structure_chunks: list[list[PassiveStructureLoop]] = []
    items = structure_inputs
    items = _progressbar(items, suffix="Structures discretized") if show_prog else items
    for inp in items:
        structure_chunks.append(PassiveStructureLoop.from_input(inp, slicer=slicer))

    # Flatten
    structure_loops: list[PassiveStructureLoop] = list(chain(*structure_chunks))

    return structure_loops



import numpy as np

import device_inductance

from pytest import approx

from interpn import MulticubicRectilinear


from . import typical_outputs, typical_outputs_stabilized_eigenmode  # Required fixture

__all__ = ["typical_outputs", "typical_outputs_stabilized_eigenmode"]


def test_coil_coil_forces(typical_outputs: device_inductance.TypicalOutputs):
    """Spot-check an entry in the coil force matrix by capitalizing on the equivalence of
    integrating I*dL and J*dV"""

    device = typical_outputs.device
    coils = device.coils

    coil_flux_density_tables = device.coil_flux_density_tables
    grids = device.grids

    for i in range(len(coils)):
        for j in range(len(coils)):
            if i == j:
                continue

            fab_mat_r, fab_mat_z = (
                device.coil_coil_forces[0][i, j],
                device.coil_coil_forces[1][i, j],
            )  # [N/A^2]

            br = coil_flux_density_tables[0][i, :, :]  # [T/A]
            bz = coil_flux_density_tables[1][i, :, :]
            br_interp = MulticubicRectilinear.new(grids, br)  # [T/A] vs. [m]
            bz_interp = MulticubicRectilinear.new(grids, bz)

            # Integral of I*cross(dL,B)/I with dL in +phi direction = 2*pi*r * nturns * (Bz, 0.0, -Br)
            length_factor = 2.0 * np.pi * coils[j].rs * coils[j].ns  # [m]-turns
            obs = [
                coils[j].rs,
                coils[j].zs,
            ]  # [m] observation points (filament locations)
            fr_interped = np.sum(length_factor * bz_interp.eval(obs))
            fz_interped = np.sum(-length_factor * br_interp.eval(obs))

            # We see a bit of interpolation error here due to the coarse grid
            # for coils that border on each other
            assert fab_mat_r == approx(fr_interped, rel=2e-2, abs=3e-6)
            assert fab_mat_z == approx(fz_interped, rel=2e-2, abs=3e-6)


def test_coil_coil_z_force_against_femm(typical_outputs: device_inductance.TypicalOutputs):
    """
    Make sure the magnitude and direction of Z-axis force between two coils
    roughly matches the result from a FEMM axisymmetric finite element model.
    See `femm_coil_test.[FEM,ans]` for FEMM problem setup and results.
    """
    device = typical_outputs.device

    coila_index = device.coil_index_dict["DV1U"]
    coilb_index = device.coil_index_dict["DV2U"]

    coila = device.coils[coila_index]
    coilb = device.coils[coilb_index]

    #
    # Leave this here for extracting the envelope to update the tests
    #

    # def coil_envelope(c: device_inductance.Coil) -> tuple[float, float, float, float]:
    #     """Coil extent only includes the filaments; this expanded extent is the input that would
    #     produce this filamentization"""
    #     rgrid, zgrid = c.grids
    #     dr, dz = np.diff(rgrid)[0], np.diff(zgrid)[0]
    #     rmin, rmax, zmin, zmax = c.extent
    #     return (rmin - dr/2, rmax + dr/2, zmin - dz/2, zmax + dz/2)
    
    # envelope_a = coil_envelope(coila)
    # envelope_b = coil_envelope(coilb)

    # print(envelope_a, envelope_b)

    # raise ValueError

    fzab = device.coil_coil_forces[1][coila_index, coilb_index]  # [N/A^2] * 1A * 1A -> [N]
    # Solved result from a FEMM model that approximately matches the coil envelope with a uniform current density
    fzab_expected = np.sum(coila.ns) * np.sum(coilb.ns) * 1.41356e-5  # [N] at 1A in each coil
    # Expect a bit of error due to FEA discretization, slight differences in envelope geometry,
    # and incomplete joint turns being distributed over the top row of filaments in device_description
    assert fzab == approx(fzab_expected, rel=0.1)


def test_circuit_coil_forces(typical_outputs: device_inductance.TypicalOutputs):
    device = typical_outputs.device
    coils = device.coils
    circuits = device.circuits
    grids = device.grids

    fr, fz = device.circuit_coil_forces
    br, bz = device.circuit_flux_density_tables

    ncoil = len(coils)
    ncirc = len(circuits)

    for i in range(ncirc):
        br_interp = MulticubicRectilinear.new(grids, br[i, :, :])  # [T/A] vs. [m]
        bz_interp = MulticubicRectilinear.new(grids, bz[i, :, :])

        circuit_coil_names = [coils[k].name for k, _ in circuits[i].coils]
        any_coils_no_self_field = any(
            [coils[k].grids is None for k, _ in circuits[i].coils]
        )

        for j in range(ncoil):
            r = coils[j].rs
            z = coils[j].zs
            n = coils[j].ns
            length_factor = 2.0 * np.pi * r * n
            fr_interped = sum(length_factor * bz_interp.eval([r, z]))
            fz_interped = sum(-length_factor * br_interp.eval([r, z]))

            if coils[j].name in circuit_coil_names and any_coils_no_self_field:
                # If this circuit-coil combination includes self-field for a coil that does not
                # have a smooth self-field calc available, then this calc will not match the test method
                pass
            elif coils[j].name in circuit_coil_names:
                # Self-field with smooth field available

                # print(i, j, circuits[i].name, coils[j].name, f"{fr[i,j]:e},{fr_interped:e}", f"{fz[i,j]:e},{fz_interped:e}")
                assert fr[i, j] == approx(fr_interped, rel=6e-2, abs=6e-6)

                # In single-coil circuits, only self-field is present, and z-force should be near zero
                # This also holds for circuits where the coils are aligned vertically.
                # All the coils in the device description fall into one of these categories as of
                # 2025-03-20, but this may need an update someday if other kinds of configurations are added.
                assert fz[i, j] == approx(0.0, abs=1e-6)
            else:
                # print(i, j, circuits[i].name, coils[j].name, f"{fr[i,j]:e},{fr_interped:e}", f"{fz[i,j]:e},{fz_interped:e}")
                assert fr[i, j] == approx(fr_interped, rel=6e-2, abs=6e-6)
                assert fz[i, j] == approx(fz_interped, rel=6e-2, abs=6e-6)


def test_structure_coil_forces(typical_outputs: device_inductance.TypicalOutputs):
    device = typical_outputs.device
    coils = device.coils

    fr, fz = device.structure_coil_forces

    # Calculate by alternative method (interpolating on field tables)
    fr_alt, fz_alt = _calc_structure_coil_forces(
        coils, device.grids, device.structure_flux_density_tables
    )

    # The interpolation method is not very good for some coils that are very closely coupled to structures,
    # so this comparison is best done in bulk across the whole population of filaments
    # and with a wide tolerance. Because the error in individual outliers is unbounded and depends primarily
    # on coincidental grid locations, the error check is implemented by limiting the number of total outlier values.
    rtol = 1e-2

    fz_rel_err = (fz - fz_alt) / fz
    fz_n_outliers = len(np.where(np.abs(fz_rel_err) > rtol)[0])

    fr_rel_err = (fr - fr_alt) / fr
    fr_n_outliers = len(np.where(np.abs(fr_rel_err) > rtol)[0])

    n_entries = fr.flatten().size
    assert fr_n_outliers < 0.05 * n_entries
    assert fz_n_outliers < 0.05 * n_entries


def test_structure_mode_coil_forces(typical_outputs: device_inductance.TypicalOutputs):
    device = typical_outputs.device
    
    fr, fz = device.structure_mode_coil_forces
    frc = fr.copy()  # Copy to mutate safely
    fzc = fz.copy()

    # Calculate by alternative method (interpolating on field tables)
    fr_alt, fz_alt = _calc_structure_mode_coil_forces(device.coils, device.grids, device.structure_mode_flux_density_tables, show_prog=False)

    # Remove VS and DV coil rows because their nearby structure is too close-coupled for this test to work well -
    # the interpolated method becomes very sensitive to how close the structures happen to be
    # to the nearest grid cell and produce intermittent test failures despite configuring gmsh to minimize
    # randomization of mesh outputs
    vs_names = [x for x in device.coil_names if "VS" in x or "DV" in x]
    vs_inds = [device.coil_index_dict[x] for x in vs_names]
    frc[:, vs_inds] = 0.0
    fzc[:, vs_inds] = 0.0
    fr_alt[:, vs_inds] = 0.0
    fz_alt[:, vs_inds] = 0.0

    # The interpolation method is not very good for some coils that are very closely coupled to structures,
    # so this comparison is best done in bulk across the whole population of filaments
    # and with a wide tolerance
    assert np.allclose(np.sum(frc, axis=0), np.sum(fr_alt, axis=0), rtol=0.2, atol=3e-6)
    assert np.allclose(np.sum(fzc, axis=0), np.sum(fz_alt, axis=0), rtol=0.2, atol=3e-6)


def test_plasma_coil_force(typical_outputs: device_inductance.TypicalOutputs, typical_outputs_stabilized_eigenmode: device_inductance.TypicalOutputs):
    device_interpolating = typical_outputs.device
    device_masked_direct = typical_outputs_stabilized_eigenmode.device
    limiter_mask = device_masked_direct.limiter_mask
    mask_inds = np.where(limiter_mask.flatten())

    # Make sure that the nonzero entries common between both methods match reasonably well
    fr_interped = device_interpolating.plasma_coil_forces[0][mask_inds, :]
    fz_interped = device_interpolating.plasma_coil_forces[1][mask_inds, :]

    fr_masked = device_masked_direct.plasma_coil_forces[0][mask_inds, :]
    fz_masked = device_masked_direct.plasma_coil_forces[1][mask_inds, :]

    assert np.allclose(fr_interped, fr_masked, rtol=2e-2, atol=1e-6)
    assert np.allclose(fz_interped, fz_masked, rtol=2e-2, atol=1e-6)



def _calc_structure_coil_forces(coils, grids, structure_flux_density_tables):
    ncoils = len(coils)
    nstruct = structure_flux_density_tables[0].shape[0]
    fr = np.zeros((nstruct, ncoils))  # [N/A^2]
    fz = np.zeros((nstruct, ncoils))

    for i in range(nstruct):
        br = structure_flux_density_tables[0][i, :, :]  # [T/A]
        bz = structure_flux_density_tables[1][i, :, :]
        br_interp = MulticubicRectilinear.new(grids, br)  # [T/A] vs. [m]
        bz_interp = MulticubicRectilinear.new(grids, bz)
        for j in range(ncoils):
            # Integral of I*cross(dL,B)/I with dL in +phi direction = 2*pi*r * nturns * (Bz, 0.0, -Br)
            length_factor = 2.0 * np.pi * coils[j].rs * coils[j].ns  # [m]-turns
            obs = [
                coils[j].rs,
                coils[j].zs,
            ]  # [m] observation points (filament locations)
            fr[i][j] = np.sum(length_factor * bz_interp.eval(obs))
            fz[i][j] = np.sum(-length_factor * br_interp.eval(obs))

    return (fr, fz)


def _calc_structure_mode_coil_forces(
    coils,
    grids,
    structure_mode_flux_density_tables,
    show_prog: bool = True,
):
    ncoils = len(coils)
    nmodes = structure_mode_flux_density_tables[0].shape[0]
    fr = np.zeros((nmodes, ncoils))  # [N/A^2]
    fz = np.zeros((nmodes, ncoils))

    for i in range(nmodes):
        br = structure_mode_flux_density_tables[0][i, :, :]  # [T/A]
        bz = structure_mode_flux_density_tables[1][i, :, :]
        br_interp = MulticubicRectilinear.new(grids, br)  # [T/A] vs. [m]
        bz_interp = MulticubicRectilinear.new(grids, bz)
        for j in range(ncoils):
            # Integral of I*cross(dL,B)/I with dL in +phi direction = 2*pi*r * nturns * (Bz, 0.0, -Br)
            length_factor = 2.0 * np.pi * coils[j].rs * coils[j].ns  # [m]-turns
            obs = [
                coils[j].rs,
                coils[j].zs,
            ]  # [m] observation points (filament locations)
            fr[i][j] = np.sum(length_factor * bz_interp.eval(obs))
            fz[i][j] = np.sum(-length_factor * br_interp.eval(obs))

    return (fr, fz)



import logging
from pathlib import Path
from rich.logging import RichHandler


def log() -> logging.Logger:
    """Get the root logger instance shared across the top level module"""
    return logging.getLogger("device_inductance")


def logger_is_set_up() -> bool:
    """Check if a rich logging handler has been set up yet"""
    return any([isinstance(x, RichHandler) for x in log().handlers])


def logger_setup(level: int, fp: Path | str | None = None):
    """Set up a logger, optionally writing to a file in addition to the terminal."""
    # Write to a file, clearing it first, if requested
    handlers = []
    if fp is not None:
        file_handler = logging.FileHandler(fp, mode="w")
        handlers.append(file_handler)

    # Always write to the terminal
    terminal_handler = RichHandler(level=level)
    handlers.append(terminal_handler)

    # Include timestamp and the origin of the message
    logging.basicConfig(
        format="%(name)s | %(asctime)s | %(message)s",
        level=level,
        handlers=handlers,
    )


def logger_setup_default():
    """Set up the default logger with WARNING level, if there is not already one set up."""
    if logger_is_set_up():
        return

    logger_setup(logging.WARNING)



"""The lowest-level discretization of conducting structure."""

from dataclasses import dataclass

import numpy as np
from shapely import Polygon, centroid
from cfsem import self_inductance_lyle6


@dataclass(frozen=True)
class PassiveStructureFilament:
    """A chunk of a cylindrically-symmetric passive conductor"""

    parent_name: str
    """Name of structure that this is associated with"""
    r: float
    """[m] radial location of filament center"""
    z: float
    """[m] axial location of filament center"""
    area: float
    """[m^2] cross-section area"""
    resistance: float
    """Ohm] loop resistance"""
    self_inductance: float
    """[H] self-inductance assuming one full loop"""
    polygon: Polygon
    """[m] Outline of mesh element"""


def _mesh_elem_to_fil(
    mesh_elem: Polygon, resistivity: float, parent_name: str
) -> PassiveStructureFilament:
    """Convert mesh element polygon to a resistive filament"""
    # Will this implicitly close? Is the mesh data ordered properly?
    area = mesh_elem.area  # [m^2]

    # Mean r,z of nodes is only really correct centroid for a proper rectangle or rhombus
    # but it should be fairly close for well-behaved shapes
    c = centroid(mesh_elem)  # [m] mean of polygon vertices
    r, z = (c.x, c.y)  # [m]
    toroidal_circumference = 2.0 * np.pi * r  # [m]
    resistance = toroidal_circumference * resistivity / area  # [Ohm]

    # Approximate as square section to estimate self-inductance
    rs, zs = mesh_elem.exterior.xy
    w = max(rs) - min(rs)
    h = max(zs) - min(zs)
    self_inductance = self_inductance_lyle6(r, w, h, n=1.0)

    return PassiveStructureFilament(
        parent_name=parent_name,
        r=r,
        z=z,
        area=area,
        resistance=resistance,
        self_inductance=self_inductance,
        polygon=mesh_elem,
    )



"""
Specialized contour-tracing algo for finding the last closed flux contour
defining the edge of a tokamak plasma.

Among other uses, this contour is needed in order to estimate the plasma's
self-inductance.
"""

from typing import Optional

import numpy as np
from numpy.typing import NDArray
from interpn import MulticubicRectilinear, MultilinearRectilinear
from scipy.optimize import minimize

from device_inductance.logging import log


def trace_contour(
    grids: tuple[NDArray, NDArray],
    psi_total: NDArray,
    start: tuple[float, float],
    maxis: tuple[float, float],
    limiter_circumference: float,
    mask_limiter: NDArray,
    ds: float = 1e-2,
    tol: float = 1e-4,
) -> Optional[tuple[NDArray, NDArray]]:
    """
    Specialized contour-tracing algo for finding the last closed flux contour
    defining the edge of a tokamak plasma.
    Among other uses, this contour is needed in order to estimate the plasma's
    self-inductance.

    Attempts to trace a closed contour starting at `start` and returns None
    if the contour crosses the limiter or does not make a full loop, or if
    the resulting contour ends up being excessively long (longer than the limiter).

    Args:
        grids: [m] 1D coordinate grids for computational domain
        psi_total: [T-m^2] 2D grid of total poloidal flux
        start: [m] Initial guess coordinates
        maxis: [m] Magnetic axis coords from O-point search
        limiter_circumference: [m] Limiter length in R-Z plane, to bound expected contour length
        mask_limiter: Float mask of limiter interior (1.0 inside, 0.0 outside)
        ds: [m] Step size. Defaults to 1e-2.
        tol: [T-m^2] Contour flux level tolerance. Defaults to 1e-4.

    Returns:
        Optional[(rpath, zpath)] contour path coordinates if a closed loop was found
    """

    # Some notes about speeding this up later if it becomes an issue
    # because this is not optimized for speed at all yet
    # * Use regular grid interpolation methods
    # * Batch interpolation calls for gradient
    # * Add jac function for rootfinder
    # * Do multilevel fixed-grid line searches from magnetic axis instead of rootfinder
    # * Do a batch check for limiter crossings once every N points
    # * Compile it if it comes to that - everything needed is available in rust, should be easy

    # Unpack
    maxis_3d = np.array([maxis[0], 0.0, maxis[1]])  # (r, phi, z)

    # Initialize limiter mask interpolator
    # This one should be linear because it's representing linear segments
    # and cubic interpolators produce over/undershoot near steps
    mask_interpolator = MultilinearRectilinear.new([x for x in grids], mask_limiter)

    # Initialize flux interpolator
    # In this case, we need a cubic function so that it can
    # properly represent a local minimum that is not exactly on a grid point.
    psi_interpolator = MulticubicRectilinear.new([x for x in grids], psi_total)

    def mask_interp_point(r, z) -> float:
        pt = [np.atleast_1d(r), np.atleast_1d(z)]
        return mask_interpolator.eval(pt)[0]

    def psi_interp_point(r, z) -> float:
        pt = [np.atleast_1d(r), np.atleast_1d(z)]
        return psi_interpolator.eval(pt)[0]

    def grad_psi(r, z, eps=1e-6) -> NDArray:
        dpsi_dr = (psi_interp_point(r + eps, z) - psi_interp_point(r - eps, z)) / (
            2.0 * eps
        )
        dpsi_dz = (psi_interp_point(r, z + eps) - psi_interp_point(r, z - eps)) / (
            2.0 * eps
        )
        return np.array((dpsi_dr, dpsi_dz))

    # Initialize
    #  Total limiter length _should_ bound the length of any reasonable flux contour
    #  since the limiter is guaranteed to be outside & typically has a less-like-a-circle shape
    #  than a typical flux contour. This is a heuristic, but a fairly dependable one
    #  as long as the flux field is sane.
    #  Preallocate output array based on length estimate and chosen step size
    n = int(np.ceil(limiter_circumference / ds)) + 1
    r = np.zeros(n)
    z = np.zeros(n)
    #  Populate starting point
    r[0], z[0] = start
    psi0 = psi_interp_point(r[0], z[0])
    # Initialize accumulated winding angle
    theta = 0.0
    for i in range(n - 2):  # access pattern is i, i+1
        # Local minimization to keep previous point on psi=psi0 contour
        sol = minimize(
            fun=lambda x: (psi_interp_point(x[0], x[1]) - psi0) ** 2,
            x0=[r[i], z[i]],
            options=dict(maxiter=100),  # type: ignore
            tol=tol,
        )
        (r[i], z[i]) = sol.x

        # Check convergence of local minimization solve.
        # If the flux field is well-behaved, it will converge,
        # and we're looking for a contour on a well-behaved region,
        # so failure to converge can be taken to mean that the
        # contour we're looking for isn't here.
        if not sol.success:
            log().error("Contour local solve did not converge")
            return None

        # Take a step along the psi=constant contour
        psi_r, psi_z = grad_psi(r[i], z[i])
        pow = 0.5  # use 0.5 for constant step sizes, less than 0.5 for adaptive steps
        step = ds / (psi_r**2 + psi_z**2) ** pow
        r[i + 1] = r[i] + psi_z * step
        z[i + 1] = z[i] - psi_r * step

        # Measure change in angle of rotation; after 2pi rotation, terminate
        # Expand vectors to (r, phi, z) coords in order to use cross product
        u = np.array([r[i + 1], 0.0, z[i + 1]]) - maxis_3d
        v = np.array([r[i], 0.0, z[i]]) - maxis_3d
        theta += np.arctan2(np.linalg.norm(np.cross(u, v)), np.dot(u, v))

        # Check if we've made a full revolution.
        # If so, finalize and return results
        if np.abs(theta) > 2.0 * np.pi:
            # Take the populated part
            r = r[: i + 2]
            z = z[: i + 2]
            # Close the contour
            r[-1] = r[0]
            z[-1] = z[0]
            return r, z  # , area

        # Check if the last point was outside the limiter after
        # its adjustment toward the psi=psi0 contour
        mask_val = mask_interp_point(r[i], z[i])
        if mask_val < 0.5:
            log().error("Contour crossed limiter")
            return None

    # We didn't make a full revolution or end up outside the limiter
    # It's possible there was a closed contour to find, but we didn't find it
    log().error("Contour exceeded maximum length without making a full revolution")
    return None



from dataclasses import dataclass
from typing import Optional

from functools import cached_property

import numpy as np
from numpy.typing import NDArray
from omas import ODS

from cfsem import (
    self_inductance_lyle6,
    flux_circular_filament,
    self_inductance_annular_ring,
    self_inductance_circular_ring_wien,
)

from .utils import solve_flux_axisymmetric, calc_flux_density_from_flux

from interpn import MulticubicRectilinear

from device_inductance.logging import log


@dataclass(frozen=True)
class CoilFilament:
    """
    A discretized element of an axisymmetric magnet.
    Self-inductance is calculated based on conductor geometry.
    """

    r: float
    """[m] radial location"""

    z: float
    """[m] z location"""

    n: float
    """[dimensionless] number of turns"""

    self_inductance: float
    """[H] scalar self-inductance of this filament"""


@dataclass(frozen=True)
class Coil:
    """An axisymmetric magnet, which may not have a rectangular cross-section"""

    name: str
    """This name should match the name used in the device description ODS"""

    resistance: float
    """[ohm] total effective resistance of the coil; for superconducting coils, this will be small"""

    self_inductance: float
    """[H] total scalar self-inductance of this coil"""

    filaments: list[CoilFilament]
    """Discretized circular filaments describing the coil's winding pattern"""

    @cached_property
    def grids(self) -> Optional[tuple[NDArray, NDArray]]:
        """Generate a set of regular r,z grids that span the coil winding pack centers exactly
        if possible, or None if the winding pack can't be represented exactly.

        Adds 4 grid cells of padding around the winding pack to deconflict the
        cells with nonzero current density from the boundary conditions of a
        flux solve.

        If only one unit cell is present on either axis, the grid will be
        expanded 1cm in either direction.
        """

        # Get coordinates with a unit cell
        unique_r = np.array(sorted(list(set([f.r for f in self.filaments]))))  # [m]
        unique_z = np.array(sorted(list(set([f.z for f in self.filaments]))))

        # Make sure there are enough unit cells to work with,
        # expanding dimensions if necessary
        if len(unique_r) == 1:
            r = unique_r[0]
            unique_r = [r - 1e-2, r, r + 1e-2]
        if len(unique_z) == 1:
            z = unique_z[0]
            unique_z = [z - 1e-2, z, z + 1e-2]
        if len(unique_r) < 2 or len(unique_z) < 2:
            log().error("Failed to expand coil grid dimensionality")
            return None

        # Check if the coordinates have regular spacing,
        # which is required to support the finite difference solve
        drs = np.diff(unique_r)
        drmean = np.mean(drs)
        if np.any(np.abs(drs - drmean) / drmean > 1e-4):
            log().warning(
                f"Coil {self.name} filaments are not on a regular grid; skipping grid for smooth self-field"
            )
            return None
        dzs = np.diff(unique_z)
        dzmean = np.mean(dzs)
        if np.any(np.abs(dzs - dzmean) / dzmean > 1e-4):
            log().warning(
                f"Coil {self.name} filaments are not on a regular grid; skipping grid for smooth self-field"
            )
            return None

        # Extend grids by a few cells outside the winding pack
        # 7 is the true minimum; 2x 4th-order finite difference patches
        # will be stacked to extract the flux then the flux density, which means
        # 7 cells see direct interaction with boundary conditions and must not
        # have nonzero current density to produce sane results; npad = 6 produces junk outputs.
        npad = 7
        nr = len(unique_r) + 2 * npad
        nz = len(unique_z) + 2 * npad
        r_pad = npad * drmean
        z_pad = npad * dzmean
        rgrid = np.linspace(unique_r[0] - r_pad, unique_r[-1] + r_pad, nr)
        zgrid = np.linspace(unique_z[0] - z_pad, unique_z[-1] + z_pad, nz)

        # Make sure the grid doesn't cross zero.
        # If this check becomes a problem, there is an alternate strategy
        # to double resolution and spread the coil's current density mapping
        # across more than one neighboring cell, but that is too much complexity
        # to implement proactively.
        if rgrid[0] < 0.0:
            log().error(f"Coil {self.name} grid for smooth self-field crossed r=0")
            return None

        return (rgrid, zgrid)

    @cached_property
    def meshes(self) -> Optional[tuple[NDArray, NDArray]]:
        """Generate a set of regular r,z meshes that span the coil winding pack centers exactly
        if possible, or None if the winding pack can't be represented exactly.

        Adds 4 grid cells of padding around the winding pack to deconflict the
        cells with nonzero current density from the boundary conditions of a
        flux solve.
        """
        grids = self.grids
        if grids is not None:
            rmesh, zmesh = np.meshgrid(*grids, indexing="ij")
            return (rmesh, zmesh)  # Unpack and repack for pyright...
        else:
            return None

    @cached_property
    def local_fields(self) -> Optional[tuple[NDArray, NDArray, NDArray]]:
        """
        Solve the local self-field flux and flux density per amp by mapping the
        coil section to a continuous current density distribution and solving the
        continuous flux field via 4th-order finite difference, then extracting the
        B-field from the flux field via 4th-order finite difference.

        Returns:
            (psi, br, bz) [Wb/A, T/A, T/A] 2D arrays of poloidal flux and flux density per amp of coil current
        """
        grids = self.grids
        meshes = self.meshes
        if grids is not None and meshes is not None:
            rgrid, zgrid = grids
            rmesh, zmesh = meshes
            dr = rgrid[1] - rgrid[0]  # [m]
            dz = zgrid[1] - zgrid[0]  # [m]
            area = dr * dz  # [m^2]

            # Map current density per amp
            jtor = np.zeros_like(meshes[0])  # [A-turns/m^2 / A]
            for f in self.filaments:
                # Get indices of location of this filament
                ri = np.argmin(np.abs(rgrid - f.r))
                zi = np.argmin(np.abs(zgrid - f.z))
                # Set current density for that unit cell
                # so that the total for the cell comes out to the
                # correct total current
                jtor[ri, zi] += f.n / area

            # Solve flux field
            psi = solve_flux_axisymmetric(grids, meshes, jtor)  # [Wb/A]

            # Extract flux density
            br, bz = calc_flux_density_from_flux(psi, rmesh, zmesh)  # [T/A]

            return psi, br, bz

        else:
            return None

    @cached_property
    def local_field_interpolators(
        self,
    ) -> Optional[
        tuple[MulticubicRectilinear, MulticubicRectilinear, MulticubicRectilinear]
    ]:
        """Build interpolators over the solved local fields, if available"""
        grids = self.grids
        local_fields = self.local_fields

        if grids is not None and local_fields is not None:
            psi, br, bz = local_fields
            grids = [x for x in grids]
            psi_interp = MulticubicRectilinear.new(grids, psi)
            br_interp = MulticubicRectilinear.new(grids, br)
            bz_interp = MulticubicRectilinear.new(grids, bz)

            return psi_interp, br_interp, bz_interp
        else:
            return None

    @cached_property
    def extent(self) -> tuple[float, float, float, float]:
        """[m] rmin, rmax, zmin, zmax extent of filament centers"""
        r = [f.r for f in self.filaments]
        z = [f.z for f in self.filaments]
        return min(r), max(r), min(z), max(z)

    @cached_property
    def rs(self) -> NDArray:
        """[m] Filament r-coordinates"""
        return np.array([f.r for f in self.filaments])

    @cached_property
    def zs(self) -> NDArray:
        """[m] Filament z-coordinates"""
        return np.array([f.z for f in self.filaments])

    @cached_property
    def ns(self) -> NDArray:
        """[dimensionless] Filament number of turns"""
        return np.array([f.n for f in self.filaments])


def _extract_coils(description: ODS) -> list[Coil]:
    """
    Extract coil filamentization and, while full geometric info is available,
    calculate self-inductance of individual filaments and of the coil as a whole.

    Because some coils are not of rectangular cross-section, and even some coils of
    rectangular cross-section do not have evenly-distributed number of turns between
    different elements, the coil's self-inductance is calculated by using the singularity
    method to calculate the mutual inductance between each pair of elements, then replacing
    the singular self-field terms with the corresponding element's estimated self-inductance.

    An approximate calc is used for the self-inductance of individual elements, which
    can't use the singularity method. The existing method adequately handles
    rectangular-section elements, but approximates each non-rectangular element as a
    square section with the same area, which may introduce some error when handling elements
    of circular, annular, or other cross-sectional geometry. More detailed handling can be
    added later, giving higher resolution for such cases.

    Args:
        description: Device geometric info in the format produced by device_description

    Raises:
        ValueError: If an un-handled type of coil element cross-sectional geometry is encountered

    Returns:
        A list of coil objects, populated with reduced geometric info and estimated self-inductances.
    """

    coils: list[Coil] = []
    for ods_coil in description["pf_active.coil"].values():
        coil_name = ods_coil["name"]
        resistance = ods_coil["resistance"]
        coil_filaments: list[CoilFilament] = []

        # Process individual elements
        for coil_elem in ods_coil["element"].values():
            geom_type = coil_elem["geometry.geometry_type"]
            turns_with_sign = coil_elem["turns_with_sign"]  # [dimensionless]
            n = abs(turns_with_sign)  # Circuit definition is responsible for sign

            # Approximate the self-inductance of the individual elements
            # as rectangular sections, solid rings, or annular rings
            # depending on geometry type id.
            if geom_type == 5:
                # Annular section (or, with zero inner radius, solid circular)
                r = coil_elem["geometry.annulus.r"]  # [m]
                z = coil_elem["geometry.annulus.z"]  # [m]
                ri = coil_elem["geometry.annulus.radius_inner"]  # [m]
                ro = coil_elem["geometry.annulus.radius_outer"]  # [m]

                if ri > 1e-4:
                    elem_self_inductance = self_inductance_annular_ring(
                        r, ri, ro
                    )  # [H]
                else:
                    # Use solid ring calc for small inner radius to avoid div/0
                    elem_self_inductance = self_inductance_circular_ring_wien(
                        r, ro
                    )  # [H]

            elif geom_type == 2:
                # Solid rectangular section
                r = coil_elem["geometry.rectangle.r"]  # [m]
                z = coil_elem["geometry.rectangle.z"]  # [m]
                w = coil_elem["geometry.rectangle.width"]  # [m]
                h = coil_elem["geometry.rectangle.height"]  # [m]

                elem_self_inductance = self_inductance_lyle6(r, w, h, n)  # [H]
            else:
                raise ValueError(f"Unhandled coil element geometry type: {geom_type}")

            # Store the parts we need for calculating mutual inductances
            coil_filaments.append(
                CoilFilament(r=r, z=z, n=n, self_inductance=float(elem_self_inductance))
            )

        # Calculate self-inductance of the whole coil
        coil_self_inductance = 0.0  # [H]
        elem_rs = np.array([x.r for x in coil_filaments])  # [m]
        elem_zs = np.array([x.z for x in coil_filaments])  # [m]
        elem_ns = np.array([x.n for x in coil_filaments])  # [dimensionless]
        elem_self_inductances = np.array(
            [x.self_inductance for x in coil_filaments]
        )  # [H]
        nelem = len(elem_rs)
        for i in range(nelem):
            this_r = np.array([elem_rs[i]])  # [m]
            this_z = np.array([elem_zs[i]])  # [m]
            this_n = np.array([elem_ns[i]])  # [dimensionless]

            # Use one-to-many flux calc for speed
            contribs = elem_ns * flux_circular_filament(
                ifil=this_n,  # Unit current multiplied by number of turns
                rfil=this_r,
                zfil=this_z,
                rprime=np.array(elem_rs),
                zprime=np.array(elem_zs),
            )
            # Use precalcualted self-contribution which is otherwise singular and handled separately
            contribs[i] = elem_self_inductances[i]

            # Total contribution to coil self-inductance
            turn_contrib = np.sum(contribs)
            coil_self_inductance += float(turn_contrib)

        coil = Coil(
            name=coil_name,
            resistance=resistance,  # [ohm]
            self_inductance=coil_self_inductance,  # [H]
            filaments=coil_filaments,
        )
        coils.append(coil)

    return coils
